# System modules
import copy
from functools import cache
import os
from typing import Literal, Optional

# Third party modules
import numpy as np
from pydantic import (BaseModel,
                      validator,
                      constr,
                      conlist,
                      conint,
                      confloat,
                      FilePath)
from pyFAI import load as pyfai_load
from pyFAI.multi_geometry import MultiGeometry
from pyFAI.units import AZIMUTHAL_UNITS, RADIAL_UNITS


class Detector(BaseModel):
    """Detector class to represent a single detector used in the
    experiment.

    :param prefix: Prefix of the detector in the SPEC file.
    :type prefix: str
    :param poni_file: Path to the poni file.
    :type poni_file: str
    :param mask_file: Optional path to the mask file.
    :type mask_file: str, optional
    """
    prefix: constr(strip_whitespace=True, min_length=1)
    poni_file: FilePath
    mask_file: Optional[FilePath]

    @validator('poni_file', allow_reuse=True)
    def validate_poni_file(cls, poni_file):
        """Validate the poni file by checking if it's a valid PONI
        file.

        :param poni_file: Path to the poni file.
        :type poni_file: str
        :raises ValueError: If poni_file is not a valid PONI file.
        :returns: Absolute path to the poni file.
        :rtype: str
        """
        poni_file = os.path.abspath(poni_file)
        try:
            ai = azimuthal_integrator(poni_file)
        except Exception as exc:
            raise ValueError(f'{poni_file} is not a valid PONI file') from exc
        return poni_file

    @validator('mask_file', allow_reuse=True)
    def validate_mask_file(cls, mask_file, values):
        """Validate the mask file. If a mask file is provided, it
        checks if it's a valid TIFF file.

        :param mask_file: Path to the mask file.
        :type mask_file: str or None
        :param values: A dictionary of the Detector fields.
        :type values: dict
        :raises ValueError: If mask_file is provided and it's not a
            valid TIFF file.
        :raises ValueError: If `'poni_file'` is not provided in `values`.
        :returns: Absolute path to the mask file or None.
        :rtype: str or None
        """
        if mask_file is None:
            return mask_file

        mask_file = os.path.abspath(mask_file)
        poni_file = values.get('poni_file')
        if poni_file is None:
            raise ValueError(
                'Cannot validate mask file without a PONI file.')
        try:
            mask_array = get_mask_array(mask_file, poni_file)
        except BaseException as exc:
            raise ValueError(
                f'Unable to open {mask_file} as a TIFF file') from exc
        return mask_file

    @property
    def azimuthal_integrator(self):
        """Return the azimuthal integrator associated with this
        detector.
        """
        return azimuthal_integrator(self.poni_file)

    @property
    def mask_array(self):
        """Return the mask array assocated with this detector."""
        return get_mask_array(self.mask_file, self.poni_file)


@cache
def azimuthal_integrator(poni_file:str):
    """Return the azimuthal integrator from a PONI file

    :param poni_file: path to a PONI file
    :type poni_file: str
    :return: azimuthal integrator
    :rtype: pyFAI.azimuthal_integrator.AzimuthalIntegrator
    """
    if not isinstance(poni_file, str):
        poni_file = str(poni_file)
    return pyfai_load(poni_file)


@cache
def get_mask_array(mask_file:str, poni_file:str):
    """Return a mask array associated with a detector loaded from a
    tiff file.

    :param mask_file: path to a .tiff file
    :type mask_file: str
    :param poni_file: path to a PONI file
    :type poni_file: str
    :return: the mask array loaded from `mask_file`
    :rtype: numpy.ndarray
    """
    if mask_file is not None:
        # Third party modules
        from pyspec.file.tiff import TiffFile

        if not isinstance(mask_file, str):
            mask_file = str(mask_file)

        with TiffFile(mask_file) as tiff:
            mask_array = tiff.asarray()
    else:
        mask_array = np.zeros(azimuthal_integrator(poni_file).detector.shape)
    return mask_array


class IntegrationConfig(BaseModel):
    """Class representing the configuration for a raw detector data
    integration.

    :ivar tool_type: type of integration tool; always set to
        "integration"
    :type tool_type: str, optional
    :ivar title: title of the integration
    :type title: str
    :ivar integration_type: type of integration, one of "azimuthal",
        "radial", or "cake"
    :type integration_type: str
    :ivar detectors: list of detectors used in the integration
    :type detectors: List[Detector]
    :ivar radial_units: radial units for the integration, defaults to
        `'q_A^-1'`
    :type radial_units: str, optional
    :ivar radial_min: minimum radial value for the integration range
    :type radial_min: float, optional
    :ivar radial_max: maximum radial value for the integration range
    :type radial_max: float, optional
    :ivar radial_npt: number of points in the radial range for the
        integration
    :type radial_npt: int, optional
    :ivar azimuthal_units: azimuthal units for the integration
    :type azimuthal_units: str, optional
    :ivar azimuthal_min: minimum azimuthal value for the integration
        range
    :type azimuthal_min: float, optional
    :ivar azimuthal_max: maximum azimuthal value for the integration
        range
    :type azimuthal_max: float, optional
    :ivar azimuthal_npt: number of points in the azimuthal range for
        the integration
    :type azimuthal_npt: int, optional
    :ivar error_model: error model for the integration, one of
        "poisson" or "azimuthal"
    :type error_model: str, optional
    """
    tool_type: Literal['integration'] = 'integration'
    title: constr(strip_whitespace=True, min_length=1)
    integration_type: Literal['azimuthal', 'radial', 'cake']
    detectors: conlist(item_type=Detector, min_items=1)
    radial_units: str = 'q_A^-1'
    radial_min: confloat(ge=0)
    radial_max: confloat(gt=0)
    radial_npt: conint(gt=0) = 1800
    azimuthal_units: str = 'chi_deg'
    azimuthal_min: confloat(ge=-180) = -180
    azimuthal_max: confloat(le=360) = 180
    azimuthal_npt: conint(gt=0) = 3600
    error_model: Optional[Literal['poisson', 'azimuthal']]
    sequence_index: Optional[conint(gt=0)]

    @validator('radial_units', allow_reuse=True)
    def validate_radial_units(cls, radial_units):
        """Validate the radial units for the integration.

        :param radial_units: unvalidated radial units for the
            integration
        :type radial_units: str
        :raises ValueError: if radial units are not one of the
            recognized radial units
        :return: validated radial units
        :rtype: str
        """
        if radial_units in RADIAL_UNITS.keys():
            return radial_units
        raise ValueError(
            f'Invalid radial units: {radial_units}. '
            f'Must be one of {", ".join(RADIAL_UNITS.keys())}')

    @validator('azimuthal_units', allow_reuse=True)
    def validate_azimuthal_units(cls, azimuthal_units):
        """Validate that `azimuthal_units` is one of the keys in the
        `pyFAI.units.AZIMUTHAL_UNITS` dictionary.

        :param azimuthal_units: The string representing the unit to be
            validated.
        :type azimuthal_units: str
        :raises ValueError: If `azimuthal_units` is not one of the
            keys in `pyFAI.units.AZIMUTHAL_UNITS`
        :return: The original supplied value, if is one of the keys in
            `pyFAI.units.AZIMUTHAL_UNITS`.
        :rtype: str
        """
        if azimuthal_units in AZIMUTHAL_UNITS.keys():
            return azimuthal_units
        raise ValueError(
            f'Invalid azimuthal units: {azimuthal_units}. '
            f'Must be one of {", ".join(AZIMUTHAL_UNITS.keys())}')

    def validate_range_max(range_name:str):
        """Validate the maximum value of an integration range.

        :param range_name: The name of the integration range
            (e.g. radial, azimuthal).
        :type range_name: str
        :return: The callable that performs the validation.
        :rtype: callable
        """
        def _validate_range_max(cls, range_max, values):
            """Check if the maximum value of the integration range is
            greater than its minimum value.

            :param range_max: The maximum value of the integration
                range.
            :type range_max: float
            :param values: The values of the other fields being
                validated.
            :type values: dict
            :raises ValueError: If the maximum value of the
                integration range is not greater than its minimum
                value.
            :return: The validated maximum range value
            :rtype: float
            """
            range_min = values.get(f'{range_name}_min')
            if range_min < range_max:
                return range_max
            raise ValueError(
                'Maximum value of integration range must be '
                'greater than minimum value of integration range '
                f'({range_name}_min={range_min}).')
        return _validate_range_max

    _validate_radial_max = validator(
        'radial_max',
        allow_reuse=True)(validate_range_max('radial'))
    _validate_azimuthal_max = validator(
        'azimuthal_max',
        allow_reuse=True)(validate_range_max('azimuthal'))

    def validate_for_map_config(self, map_config:BaseModel):
        """Validate the existence of the detector data file for all
        scan points in `map_config`.

        :param map_config: The `MapConfig` instance to validate
            against.
        :type map_config: MapConfig
        :raises RuntimeError: If a detector data file could not be
            found for a scan point occurring in `map_config`.
        :return: None
        :rtype: None
        """
        for detector in self.detectors:
            for scans in map_config.spec_scans:
                for scan_number in scans.scan_numbers:
                    scanparser = scans.get_scanparser(scan_number)
                    for scan_step_index in range(scanparser.spec_scan_npts):
                        # Make sure the detector data file exists for
                        # all scan points
                        try:
                            detector_data_file = \
                                scanparser.get_detector_data_file(
                                    detector.prefix, scan_step_index)
                        except Exception as exc:
                            raise RuntimeError(
                                'Could not find data file for detector prefix '
                                f'{detector.prefix} '
                                f'on scan number {scan_number} '
                                f'in spec file {scans.spec_file}') from exc

    def get_azimuthal_adjustments(self):
        """To enable a continuous range of integration in the
        azimuthal direction for radial and cake integration, obtain
        adjusted values for this `IntegrationConfig`'s `azimuthal_min`
        and `azimuthal_max` values, the angle amount by which those
        values were adjusted, and the proper location of the
        discontinuity in the azimuthal direction.

        :return: Adjusted chi_min, adjusted chi_max, chi_offset,
            chi_discontinuity
        :rtype: tuple[float,float,float,float]
        """
        return get_azimuthal_adjustments(self.azimuthal_min,
                                         self.azimuthal_max)

    def get_azimuthal_integrators(self):
        """Get a list of `AzimuthalIntegrator`s that correspond to the
        detector configurations in this instance of
        `IntegrationConfig`.

        The returned `AzimuthalIntegrator`s are (if need be)
        artificially rotated in the azimuthal direction to achieve a
        continuous range of integration in the azimuthal direction.

        :returns: A list of `AzimuthalIntegrator`s appropriate for use
            by this `IntegrationConfig` tool
        :rtype: list[pyFAI.azimuthalIntegrator.AzimuthalIntegrator]
        """
        chi_offset = self.get_azimuthal_adjustments()[2]
        return get_azimuthal_integrators(
            tuple([detector.poni_file for detector in self.detectors]),
            chi_offset=chi_offset)

    def get_multi_geometry_integrator(self):
        """Get a `MultiGeometry` integrator suitable for use by this
        instance of `IntegrationConfig`.

        :return: A `MultiGeometry` integrator
        :rtype: pyFAI.multi_geometry.MultiGeometry
        """
        poni_files = tuple([detector.poni_file for detector in self.detectors])
        radial_range = (self.radial_min, self.radial_max)
        azimuthal_range = (self.azimuthal_min, self.azimuthal_max)
        return get_multi_geometry_integrator(poni_files, self.radial_units,
                                             radial_range, azimuthal_range)

    def get_azimuthally_integrated_data(self,
                                        spec_scans:BaseModel,
                                        scan_number:int,
                                        scan_step_index:int):
        """Return azimuthally-integrated data for the scan step
        specified.

        :param spec_scans: An instance of `SpecScans` containing the
            scan step requested.
        :type spec_scans: SpecScans
        :param scan_number: The number of the scan containing the scan
            step requested.
        :type scan_number: int
        :param scan_step_index: The index of the scan step requested.
        :type scan_step_index: int
        :return: A 1D array of azimuthally-integrated raw detector
            intensities.
        :rtype: np.ndarray
        """
        detector_data = spec_scans.get_detector_data(self.detectors,
                                                     scan_number,
                                                     scan_step_index)
        integrator = self.get_multi_geometry_integrator()
        lst_mask = [detector.mask_array for detector in self.detectors]
        result = integrator.integrate1d(detector_data,
                                        lst_mask=lst_mask,
                                        npt=self.radial_npt,
                                        error_model=self.error_model)
        if result.sigma is None:
            return result.intensity
        return result.intensity, result.sigma

    def get_radially_integrated_data(self,
                                     spec_scans:BaseModel,
                                     scan_number:int,
                                     scan_step_index:int):
        """Return radially-integrated data for the scan step
        specified.

        :param spec_scans: An instance of `SpecScans` containing the
            scan step requested.
        :type spec_scans: SpecScans
        :param scan_number: The number of the scan containing the scan
            step requested.
        :type scan_number: int
        :param scan_step_index: The index of the scan step requested.
        :type scan_step_index: int
        :return: A 1D array of radially-integrated raw detector
            intensities.
        :rtype: np.ndarray
        """
        # Handle idiosyncracies of azimuthal ranges in pyFAI Adjust
        # chi ranges to get a continuous range of iintegrated data
        chi_min, chi_max, *adjust = self.get_azimuthal_adjustments()
        # Perform radial integration on a detector-by-detector basis.
        intensity_each_detector = []
        variance_each_detector = []
        integrators = self.get_azimuthal_integrators()
        for integrator, detector in zip(integrators, self.detectors):
            detector_data = spec_scans.get_detector_data(
                [detector], scan_number, scan_step_index)[0]
            result = integrator.integrate_radial(
                detector_data,
                self.azimuthal_npt,
                unit=self.azimuthal_units,
                azimuth_range=(chi_min, chi_max),
                radial_unit=self.radial_units,
                radial_range=(self.radial_min, self.radial_max),
                mask=detector.mask_array)  # , error_model=self.error_model)
            intensity_each_detector.append(result.intensity)
            if result.sigma is not None:
                variance_each_detector.append(result.sigma**2)
        # Add the individual detectors' integrated intensities
        # together
        intensity = np.nansum(intensity_each_detector, axis=0)
        # Ignore data at values of chi for which there was no data
        intensity = np.where(intensity == 0, np.nan, intensity)
        if len(intensity_each_detector) != len(variance_each_detector):
            return intensity

        # Get the standard deviation of the summed detectors'
        # intensities
        sigma = np.sqrt(np.nansum(variance_each_detector, axis=0))
        return intensity, sigma

    def get_cake_integrated_data(self,
                                 spec_scans:BaseModel,
                                 scan_number:int,
                                 scan_step_index:int):
        """Return cake-integrated data for the scan step specified.

        :param spec_scans: An instance of `SpecScans` containing the
            scan step requested.
        :type spec_scans: SpecScans
        :param scan_number: The number of the scan containing the scan
            step requested.
        :type scan_number: int
        :param scan_step_index: The index of the scan step requested.
        :type scan_step_index: int
        :return: A 2D array of cake-integrated raw detector
            intensities.
        :rtype: np.ndarray
        """
        detector_data = spec_scans.get_detector_data(
            self.detectors, scan_number, scan_step_index)
        integrator = self.get_multi_geometry_integrator()
        lst_mask = [detector.mask_array for detector in self.detectors]
        result = integrator.integrate2d(
            detector_data,
            lst_mask=lst_mask,
            npt_rad=self.radial_npt,
            npt_azim=self.azimuthal_npt,
            method='bbox',
            error_model=self.error_model)
        if result.sigma is None:
            return result.intensity
        return result.intensity, result.sigma

    def get_integrated_data(self,
                            spec_scans:BaseModel,
                            scan_number:int,
                            scan_step_index:int):
        """Return integrated data for the scan step specified.

        :param spec_scans: An instance of `SpecScans` containing the
            scan step requested.
        :type spec_scans: SpecScans
        :param scan_number: The number of the scan containing the scan
            step requested.
        :type scan_number: int
        :param scan_step_index: The index of the scan step requested.
        :type scan_step_index: int
        :return: An array of integrated raw detector intensities.
        :rtype: np.ndarray
        """
        if self.integration_type == 'azimuthal':
            return self.get_azimuthally_integrated_data(spec_scans,
                                                        scan_number,
                                                        scan_step_index)
        if self.integration_type == 'radial':
            return self.get_radially_integrated_data(spec_scans,
                                                     scan_number,
                                                     scan_step_index)
        if self.integration_type == 'cake':
            return self.get_cake_integrated_data(spec_scans,
                                                 scan_number,
                                                 scan_step_index)
        return None

    @property
    def integrated_data_coordinates(self):
        """Return a dictionary of coordinate arrays for navigating the
        dimension(s) of the integrated data produced by this instance
        of `IntegrationConfig`.

        :return: A dictionary with either one or two keys: 'azimuthal'
            and/or 'radial', each of which points to a 1-D `numpy`
            array of coordinate values.
        :rtype: dict[str,np.ndarray]
        """
        if self.integration_type == 'azimuthal':
            return get_integrated_data_coordinates(
                radial_range=(self.radial_min, self.radial_max),
                radial_npt=self.radial_npt)
        if self.integration_type == 'radial':
            return get_integrated_data_coordinates(
                azimuthal_range=(self.azimuthal_min, self.azimuthal_max),
                azimuthal_npt=self.azimuthal_npt)
        if self.integration_type == 'cake':
            return get_integrated_data_coordinates(
                radial_range=(self.radial_min, self.radial_max),
                radial_npt=self.radial_npt,
                azimuthal_range=(self.azimuthal_min, self.azimuthal_max),
                azimuthal_npt=self.azimuthal_npt)
        return None

    @property
    def integrated_data_dims(self):
        """Return a tuple of the coordinate labels for the integrated
        data produced by this instance of `IntegrationConfig`.
        """
        directions = list(self.integrated_data_coordinates.keys())
        dim_names = [getattr(self, f'{direction}_units')
                     for direction in directions]
        return dim_names

    @property
    def integrated_data_shape(self):
        """Return a tuple representing the shape of the integrated
        data produced by this instance of `IntegrationConfig` for a
        single scan step.
        """
        return tuple([len(coordinate_values)
                      for coordinate_name, coordinate_values
                      in self.integrated_data_coordinates.items()])


@cache
def get_azimuthal_adjustments(chi_min:float, chi_max:float):
    """Fix chi discontinuity at 180 degrees and return the adjusted
    chi range, offset, and discontinuty.

    If the discontinuity is crossed, obtain the offset to artificially
    rotate detectors to achieve a continuous azimuthal integration
    range.

    :param chi_min: The minimum value of the azimuthal range.
    :type chi_min: float
    :param chi_max: The maximum value of the azimuthal range.
    :type chi_max: float
    :return: The following four values: the adjusted minimum value of
        the azimuthal range, the adjusted maximum value of the
        azimuthal range, the value by which the chi angle was
        adjusted, the position of the chi discontinuity.
    """
    # Fix chi discontinuity at 180 degrees for now.
    chi_disc = 180
    # If the discontinuity is crossed, artificially rotate the
    # detectors to achieve a continuous azimuthal integration range
    if chi_min < chi_disc and chi_max > chi_disc:
        chi_offset = chi_max - chi_disc
    else:
        chi_offset = 0
    return chi_min-chi_offset, chi_max-chi_offset, chi_offset, chi_disc


@cache
def get_azimuthal_integrators(poni_files:tuple, chi_offset=0):
    """Return a list of `AzimuthalIntegrator` objects generated from
    PONI files.

    :param poni_files: Tuple of strings, each string being a path to a
        PONI file.
    :type poni_files: tuple
    :param chi_offset: The angle in degrees by which the
        `AzimuthalIntegrator` objects will be rotated, defaults to 0.
    :type chi_offset: float, optional
    :return: List of `AzimuthalIntegrator` objects
    :rtype: list[pyFAI.azimuthalIntegrator.AzimuthalIntegrator]
    """
    ais = []
    for poni_file in poni_files:
        ai = copy.deepcopy(azimuthal_integrator(poni_file))
        ai.rot3 += chi_offset * np.pi/180
        ais.append(ai)
    return ais


@cache
def get_multi_geometry_integrator(poni_files:tuple, radial_unit:str,
                                  radial_range:tuple, azimuthal_range:tuple):
    """Return a `MultiGeometry` instance that can be used for
    azimuthal or cake integration.

    :param poni_files: Tuple of PONI files that describe the detectors
        to be integrated.
    :type poni_files: tuple
    :param radial_unit: Unit to use for radial integration range.
    :type radial_unit: str
    :param radial_range: Tuple describing the range for radial integration.
    :type radial_range: tuple[float,float]
    :param azimuthal_range:Tuple describing the range for azimuthal
        integration.
    :type azimuthal_range: tuple[float,float]
    :return: `MultiGeometry` instance that can be used for azimuthal
        or cake integration.
    :rtype: pyFAI.multi_geometry.MultiGeometry
    """
    chi_min, chi_max, chi_offset, chi_disc = \
        get_azimuthal_adjustments(*azimuthal_range)
    ais = copy.deepcopy(get_azimuthal_integrators(poni_files,
                                                  chi_offset=chi_offset))
    multi_geometry = MultiGeometry(
        ais,
        unit=radial_unit,
        radial_range=radial_range,
        azimuth_range=(chi_min, chi_max),
        wavelength=sum([ai.wavelength for ai in ais])/len(ais),
        chi_disc=chi_disc)
    return multi_geometry


@cache
def get_integrated_data_coordinates(azimuthal_range:tuple = None,
                                    azimuthal_npt:int = None,
                                    radial_range:tuple = None,
                                    radial_npt:int = None):
    """Return a dictionary of coordinate arrays for the specified
    radial and/or azimuthal integration ranges.

    :param azimuthal_range: Tuple specifying the range of azimuthal
        angles over which to generate coordinates, in the format (min,
        max), defaults to None.
    :type azimuthal_range: tuple[float,float], optional
    :param azimuthal_npt: Number of azimuthal coordinate points to
        generate, defaults to None.
    :type azimuthal_npt: int, optional
    :param radial_range: Tuple specifying the range of radial
        distances over which to generate coordinates, in the format
        (min, max), defaults to None.
    :type radial_range: tuple[float,float], optional
    :param radial_npt: Number of radial coordinate points to generate,
        defaults to None.
    :type radial_npt: int, optional
    :return: A dictionary with either one or two keys: 'azimuthal'
        and/or 'radial', each of which points to a 1-D `numpy` array
        of coordinate values.
    :rtype: dict[str,np.ndarray]
    """
    integrated_data_coordinates = {}
    if azimuthal_range is not None and azimuthal_npt is not None:
        integrated_data_coordinates['azimuthal'] = np.linspace(
            *azimuthal_range, azimuthal_npt)
    if radial_range is not None and radial_npt is not None:
        integrated_data_coordinates['radial'] = np.linspace(
            *radial_range, radial_npt)
    return integrated_data_coordinates



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : processor.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Processor module

Define a generic `Processor` object.
"""

# system modules
import argparse
from inspect import getfullargspec
import logging
from sys import modules
from time import time

# local modules
from CHAP.pipeline import PipelineItem

class Processor(PipelineItem):
    """Generic data processor.

    The job of any `Processor` in a `Pipeline` is to receive data
    returned by the previous `PipelineItem`, process it in some way,
    and return the result for the next `PipelineItem` to use as input.
    """

    def process(self, data):
        """Extract the contents of the input data, add a string to it,
        and return the amended value.

        :param data: input data
        :return: processed data
        """
        # If needed, extract data from a returned value of Reader.read
        if isinstance(data, list):
            if all(isinstance(d, dict) for d in data):
                data = data[0]['data']
        if data is None:
            return []
        # process operation is a simple string concatenation
        data += "process part\n"
        # and we return data back to pipeline
        return data


class OptionParser():
    """User based option parser"""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--data', action='store',
            dest='data', default='', help='Input data')
        self.parser.add_argument(
            '--processor', action='store',
            dest='processor', default='Processor', help='Processor class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function"""

    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.processor
    try:
        processor_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported processor {cls_name}')
        raise

    processor = processor_cls()
    processor.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    processor.logger.addHandler(log_handler)
    data = processor.process(opts.data)

    print(f'Processor {processor} operates on data {data}')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : processor.py
Author     : Keara Soloway, Rolf Verberg
Description: Module for Processors used only by EDD experiments
"""

# system modules
from json import dumps
import os

# third party modules
import numpy as np

# local modules
from CHAP.processor import Processor


class DiffractionVolumeLengthProcessor(Processor):
    """A Processor using a steel foil raster scan to calculate the
    length of the diffraction volume for an EDD setup.
    """

    def process(self,
                data,
                config=None,
                save_figures=False,
                outputdir='.',
                inputdir='.',
                interactive=False):
        """Return calculated value of the DV length.

        :param data: input configuration for the raw scan data & DVL
            calculation procedure.
        :type data: list[PipelineData]
        :param config: initialization parameters for an instance of
            CHAP.edd.models.DiffractionVolumeLengthConfig, defaults to
            None
        :type config: dict, optional
        :param save_figures: save .pngs of plots for checking inputs &
            outputs of this Processor, defaults to False
        :type save_figures: bool, optional
        :param outputdir: directory to which any output figures will
            be saved, defaults to '.'
        :type outputdir: str, optional
        :param inputdir: input directory, used only if files in the
            input configuration are not absolute paths.
            be saved, defaults to '.'
        :type inputdir: str, optional
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: complete DVL configuraiton dictionary
        :rtype: dict
        """

        try:
            dvl_config = self.get_config(
                data, 'edd.models.DiffractionVolumeLengthConfig',
                inputdir=inputdir)
        except Exception as data_exc:
            self.logger.info('No valid DVL config in input pipeline data, '
                             + 'using config parameter instead.')
            try:
                from CHAP.edd.models import DiffractionVolumeLengthConfig
                dvl_config = DiffractionVolumeLengthConfig(
                    **config, inputdir=inputdir)
            except Exception as dict_exc:
                self.logger.error('Could not get a valid DVL config')
                raise RuntimeError from dict_exc

        for detector in dvl_config.detectors:
            dvl = self.measure_dvl(dvl_config, detector,
                                   save_figures=save_figures,
                                   interactive=interactive,
                                   outputdir=outputdir)
            detector.dvl_measured = dvl

        return dvl_config.dict()

    def measure_dvl(self,
                    dvl_config,
                    detector,
                    save_figures=False,
                    outputdir='.',
                    interactive=False):
        """Return a measured value for the length of the diffraction
        volume. Use the iron foil raster scan data provided in
        `dvl_config` and fit a gaussian to the sum of all MCA channel
        counts vs scanned motor position in the raster scan. The
        computed diffraction volume length is approximately equal to
        the standard deviation of the fitted peak.

        :param dvl_config: configuration for the DVL calculation
            procedure
        :type dvl_config: DiffractionVolumeLengthConfig
        :param detector: A single MCA detector element configuration
        :type detector: CHAP.edd.models.MCAElementDiffractionVolumeLengthConfig
        :param save_figures: save .pngs of plots for checking inputs &
            outputs of this Processor, defaults to False
        :type save_figures: bool, optional
        :param outputdir: directory to which any output figures will
            be saved, defaults to '.'
        :type outputdir: str, optional
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: calculated diffraction volume length
        :rtype: float
        """

        from CHAP.utils.fit import Fit
        from CHAP.utils.general import draw_mask_1d

        # Get raw MCA data from raster scan
        mca_data = dvl_config.mca_data(detector)

        # Interactively set mask, if needed & possible.
        if interactive or save_figures:
            self.logger.info(
                'Interactively select a mask in the matplotlib figure')
            mask, include_bin_ranges, figure = draw_mask_1d(
                np.sum(mca_data, axis=0),
                xdata = np.arange(detector.num_bins),
                current_index_ranges=detector.include_bin_ranges,
                label='sum of MCA spectra over all scan points',
                title='Click and drag to select ranges of MCA data to\n'
                + 'include when measuring the diffraction volume length.',
                xlabel='MCA channel (index)',
                ylabel='MCA intensity (counts)',
                test_mode=not interactive,
                return_figure=True
            )
            detector.include_bin_ranges = include_bin_ranges
            self.logger.debug('Mask selected. Including detector bin ranges: '
                              + str(detector.include_bin_ranges))
            if save_figures:
                figure.savefig(os.path.join(
                    outputdir, f'{detector.detector_name}_dvl_mask.png'))
            import matplotlib.pyplot as plt
            plt.close()
        if detector.include_bin_ranges is None:
            raise ValueError(
                'No value provided for include_bin_ranges. '
                + 'Provide them in the Diffraction Volume Length '
                + 'Measurement Configuration, or re-run the pipeline '
                + 'with the --interactive flag.')

        # Reduce the raw MCA data in 3 ways:
        # 1) sum of intensities in all detector bins
        # 2) max of intensities in detector bins after mask is applied
        # 3) sum of intensities in detector bins after mask is applied
        unmasked_sum = np.sum(mca_data, axis=1)
        mask = detector.mca_mask()
        masked_mca_data = np.empty(
            (mca_data.shape[0], *mca_data[0][mask].shape))
        for i in range(mca_data.shape[0]):
            masked_mca_data[i] = mca_data[i][mask]
        masked_max = np.amax(masked_mca_data, axis=1)
        masked_sum = np.sum(masked_mca_data, axis=1)

        # Find the motor position corresponding roughly to the center
        # of the diffraction volume
        scanned_vals = dvl_config.scanned_vals
        scan_center = np.sum(scanned_vals * masked_sum) / np.sum(masked_sum)
        x = scanned_vals - scan_center

        # "Normalize" the masked summed data and fit a gaussian to it
        y = (masked_sum - min(masked_sum)) / max(masked_sum)
        fit = Fit.fit_data(y, 'gaussian', x=x, normalize=False)

        # Calculate / manually select diffraction volume length
        dvl = fit.best_values['sigma'] * detector.sigma_to_dvl_factor
        if detector.measurement_mode == 'manual':
            if interactive:
                mask, dvl_bounds = draw_mask_1d(
                    y, xdata=x,
                    label='total (masked & normalized)',
                    ref_data=[
                        ((x, fit.best_fit),
                         {'label': 'gaussian fit (to total)'}),
                        ((x, masked_max / max(masked_max)),
                         {'label': 'maximum (masked)'}),
                        ((x, unmasked_sum / max(unmasked_sum)),
                         {'label': 'total (unmasked)'})
                    ],
                    num_index_ranges_max=1,
                    title=('Click and drag to indicate the\n'
                           + 'boundary of the diffraction volume'),
                    xlabel=(dvl_config.scanned_dim_lbl
                            + ' (offset from scan "center")'),
                    ylabel='MCA intensity (normalized)')
                dvl_bounds = dvl_bounds[0]
                dvl = abs(x[dvl_bounds[1]] - x[dvl_bounds[0]])
            else:
                self.logger.warning(
                    'Cannot manually indicate DVL when running CHAP '
                    + 'non-interactively. '
                    + 'Using default DVL calcluation instead.')

        if interactive or save_figures:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots()
            ax.set_title(f'Diffraction Volume ({detector.detector_name})')
            ax.set_xlabel(dvl_config.scanned_dim_lbl \
                          + ' (offset from scan "center")')
            ax.set_ylabel('MCA intensity (normalized)')
            ax.plot(x, y, label='total (masked & normalized)')
            ax.plot(x, fit.best_fit, label='gaussian fit (to total)')
            ax.plot(x, masked_max / max(masked_max),
                    label='maximum (masked)')
            ax.plot(x, unmasked_sum / max(unmasked_sum),
                    label='total (unmasked)')
            ax.axvspan(-dvl / 2., dvl / 2.,
                       color='gray', alpha=0.5,
                       label='diffraction volume'
                       + f' ({detector.measurement_mode})')
            ax.legend()

            if save_figures:
                figfile = os.path.join(outputdir,
                                       f'{detector.detector_name}_dvl.png')
                plt.savefig(figfile)
                self.logger.info(f'Saved figure to {figfile}')
            if interactive:
                plt.show()

        return dvl

class MCACeriaCalibrationProcessor(Processor):
    """A Processor using a CeO2 scan to obtain tuned values for the
    bragg diffraction angle and linear correction parameters for MCA
    channel energies for an EDD experimental setup.
    """

    def process(self,
                data,
                config=None,
                save_figures=False,
                outputdir='.',
                inputdir='.',
                interactive=False):
        """Return tuned values for 2&theta and linear correction
        parameters for the MCA channel energies.

        :param data: input configuration for the raw data & tuning
            procedure
        :type data: list[dict[str,object]]
        :param config: initialization parameters for an instance of
            CHAP.edd.models.MCACeriaCalibrationConfig, defaults to
            None
        :type config: dict, optional
        :param save_figures: save .pngs of plots for checking inputs &
            outputs of this Processor, defaults to False
        :type save_figures: bool, optional
        :param outputdir: directory to which any output figures will
            be saved, defaults to '.'
        :type outputdir: str, optional
        :param inputdir: input directory, used only if files in the
            input configuration are not absolute paths.
            be saved, defaults to '.'
        :type inputdir: str, optional
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: original configuration dictionary with tuned values
            added
        :rtype: dict[str,float]
        """

        try:
            calibration_config = self.get_config(
                data, 'edd.models.MCACeriaCalibrationConfig',
                inputdir=inputdir)
        except Exception as data_exc:
            self.logger.info('No valid calibration config in input pipeline '
                             + 'data, using config parameter instead.')
            try:
                from CHAP.edd.models import MCACeriaCalibrationConfig
                calibration_config = MCACeriaCalibrationConfig(
                    **config, inputdir=inputdir)
            except Exception as dict_exc:
                raise RuntimeError from dict_exc

        for detector in calibration_config.detectors:
            tth, slope, intercept = self.calibrate(
                calibration_config, detector,
                save_figures=save_figures,
                interactive=interactive, outputdir=outputdir)

            detector.tth_calibrated = tth
            detector.slope_calibrated = slope
            detector.intercept_calibrated = intercept

        return calibration_config.dict()

    def calibrate(self,
                  calibration_config,
                  detector,
                  save_figures=False,
                  outputdir='.',
                  interactive=False):
        """Iteratively calibrate 2&theta by fitting selected peaks of
        an MCA spectrum until the computed strain is sufficiently
        small. Use the fitted peak locations to determine linear
        correction parameters for the MCA's channel energies.

        :param calibration_config: object configuring the CeO2
            calibration procedure
        :type calibration_config: MCACeriaCalibrationConfig
        :param detector: a single MCA detector element configuration
        :type detector: CHAP.edd.models.MCAElementCalibrationConfig
        :param save_figures: save .pngs of plots for checking inputs &
            outputs of this Processor, defaults to False
        :type save_figures: bool, optional
        :param outputdir: directory to which any output figures will
            be saved, defaults to '.'
        :type outputdir: str, optional
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: calibrated values of 2&theta and linear correction
            parameters for MCA channel energies : tth, slope,
            intercept
        :rtype: float, float, float
        """
        from CHAP.edd.utils import hc
        from CHAP.utils.fit import Fit

        # Collect raw MCA data of interest
        mca_data = calibration_config.mca_data(detector)
        mca_bin_energies = np.arange(0, detector.num_bins) \
            * (detector.max_energy_kev / detector.num_bins)

        if interactive:
            # Interactively adjust initial tth guess
            from CHAP.edd.utils import select_tth_initial_guess
            select_tth_initial_guess(detector, calibration_config.material,
                                     mca_data, mca_bin_energies)
        self.logger.debug(f'tth_initial_guess = {detector.tth_initial_guess}')

        # Mask out the corrected MCA data for fitting
        if interactive:
            from CHAP.utils.general import draw_mask_1d
            self.logger.info(
                'Interactively select a mask in the matplotlib figure')
            mask, include_bin_ranges = draw_mask_1d(
                mca_data,
                xdata=mca_bin_energies,
                current_index_ranges=detector.include_bin_ranges,
                title='Click and drag to select ranges of Ceria'
                +' calibration data to include',
                xlabel='MCA channel energy (keV)',
                ylabel='MCA intensity (counts)')
            detector.include_bin_ranges = include_bin_ranges
            self.logger.debug('Mask selected. Including detector bin ranges: '
                              + str(detector.include_bin_ranges))
        if detector.include_bin_ranges is None:
            raise ValueError(
                'No value provided for include_bin_ranges. '
                'Provide them in the MCA Ceria Calibration Configuration, '
                'or re-run the pipeline with the --interactive flag.')
        mca_mask = detector.mca_mask()
        fit_mca_energies = mca_bin_energies[mca_mask]
        fit_mca_intensities = mca_data[mca_mask]

        # Correct raw MCA data for variable flux at different energies
        flux_correct = \
            calibration_config.flux_correction_interpolation_function()
        mca_intensity_weights = flux_correct(fit_mca_energies)
        fit_mca_intensities = fit_mca_intensities / mca_intensity_weights

        # Get the HKLs and lattice spacings that will be used for
        # fitting
        tth = detector.tth_initial_guess
        if interactive or save_figures:
            import matplotlib.pyplot as plt
            from CHAP.edd.utils import select_hkls
            fig = select_hkls(detector, [calibration_config.material], tth,
                              mca_data, mca_bin_energies, interactive)
            if save_figures:
                fig.savefig(os.path.join(
                    outputdir,
                    f'{detector.detector_name}_calibration_hkls.png'))
            plt.close()
        self.logger.debug(f'HKLs selected: {detector.fit_hkls}')
        if detector.fit_hkls is None:
            raise ValueError(
                'No value provided for fit_hkls. Provide them in '
                'the detector\'s MCA Ceria Calibration Configuration, or'
                ' re-run the pipeline with the --interactive flag.')
        fit_hkls, fit_ds = detector.fit_ds(calibration_config.material)
        c_1 = fit_hkls[:,0]**2 + fit_hkls[:,1]**2 + fit_hkls[:,2]**2

        for iter_i in range(calibration_config.max_iter):
            self.logger.debug(f'Tuning tth: iteration no. {iter_i}, '
                              + f'starting tth value = {tth} ')

            # Perform the uniform fit first

            # Get expected peak energy locations for this iteration's
            # starting value of tth
            fit_lambda = 2.0*fit_ds*np.sin(0.5*np.radians(tth))
            fit_E0 = hc / fit_lambda

            # Run the uniform fit
            uniform_fit = Fit(fit_mca_intensities, x=fit_mca_energies)
            uniform_fit.create_multipeak_model(fit_E0, fit_type='uniform')
            uniform_fit.fit()

            # Extract values of interest from the best values for the
            # uniform fit parameters
            uniform_fit_centers = [
                uniform_fit.best_values[f'peak{i+1}_center']
                for i in range(len(detector.fit_hkls))]
            uniform_a = uniform_fit.best_values['scale_factor']
            uniform_strain = np.log(
                (uniform_a
                 / calibration_config.material.lattice_parameters)) # CeO2 is cubic, so this is fine here.

            # Next, perform the unconstrained fit

            # Use the peak locations found in the uniform fit as the
            # initial guesses for peak locations in the unconstrained
            # fit
            unconstrained_fit = Fit(fit_mca_intensities, x=fit_mca_energies)
            unconstrained_fit.create_multipeak_model(
                uniform_fit_centers, fit_type='unconstrained')
            unconstrained_fit.fit()

            # Extract values of interest from the best values for the
            # unconstrained fit parameters
            unconstrained_fit_centers = np.array(
                [unconstrained_fit.best_values[f'peak{i+1}_center']
                 for i in range(len(detector.fit_hkls))])
            unconstrained_a = 0.5*hc*np.sqrt(c_1) \
                / (unconstrained_fit_centers*abs(np.sin(0.5*np.radians(tth))))
            unconstrained_strains = np.log(
                (unconstrained_a
                 / calibration_config.material.lattice_parameters))
            unconstrained_strain = np.mean(unconstrained_strains)
            unconstrained_tth = tth * (1.0 + unconstrained_strain)

            # Update tth for the next iteration of tuning
            prev_tth = tth
            tth = unconstrained_tth

            # Stop tuning tth at this iteration if differences are
            # small enough
            if abs(tth - prev_tth) < calibration_config.tune_tth_tol:
                break

        # Fit line to expected / computed peak locations from the last
        # unconstrained fit.
        fit = Fit.fit_data(
            fit_E0,
            'linear',
            x=unconstrained_fit_centers,
            nan_policy='omit')
        slope = fit.best_values['slope']
        intercept = fit.best_values['intercept']

        if interactive or save_figures:
            fig, axs = plt.subplots(2, 2, sharex='all', figsize=(11, 8.5))

            # Upper left axes: Input data & best fits
            axs[0,0].set_title('Ceria Calibration Fits')
            axs[0,0].set_xlabel('Energy (keV)')
            axs[0,0].set_ylabel('Intensity (a.u)')
            for i, hkl_E in enumerate(fit_E0):
                # KLS: annotate indicated HKLs w millier indices
                axs[0,0].axvline(hkl_E, color='k', linestyle='--')
                axs[0,0].text(hkl_E, 1, str(fit_hkls[i])[1:-1],
                              ha='right', va='top', rotation=90,
                              transform=axs[0,0].get_xaxis_transform())
            axs[0,0].plot(fit_mca_energies, uniform_fit.best_fit,
                        label='Single Strain')
            axs[0,0].plot(fit_mca_energies, unconstrained_fit.best_fit,
                        label='Unconstrained')
            #axs[0,0].plot(fit_mca_energies, MISSING?, label='least squares')
            axs[0,0].plot(fit_mca_energies, fit_mca_intensities,
                        label='Flux-Corrected & Masked MCA Data')
            axs[0,0].legend()

            # Lower left axes: fit residuals
            axs[1,0].set_title('Fit Residuals')
            axs[1,0].set_xlabel('Energy (keV)')
            axs[1,0].set_ylabel('Residual (a.u)')
            axs[1,0].plot(fit_mca_energies,
                          uniform_fit.residual,
                          label='Single Strain')
            axs[1,0].plot(fit_mca_energies,
                          unconstrained_fit.residual,
                          label='Unconstrained')
            axs[1,0].legend()

            # Upper right axes: E vs strain for each fit
            axs[0,1].set_title('HKL Energy vs. Microstrain')
            axs[0,1].set_xlabel('Energy (keV)')
            axs[0,1].set_ylabel('Strain (\u03BC\u03B5)')
            axs[0,1].axhline(uniform_strain * 1e6,
                             linestyle='--', label='Single Strain')
            axs[0,1].plot(fit_E0, unconstrained_strains * 1e6,
                          color='C1', marker='s', label='Unconstrained')
            axs[0,1].axhline(unconstrained_strain * 1e6,
                             color='C1', linestyle='--',
                             label='Unconstrained: Unweighted Mean')
            axs[0,1].legend()

            # Lower right axes: theoretical HKL E vs fit HKL E for
            # each fit
            axs[1,1].set_title('Theoretical vs. Fit HKL Energies')
            axs[1,1].set_xlabel('Energy (keV)')
            axs[1,1].set_ylabel('Energy (keV)')
            axs[1,1].plot(fit_E0, uniform_fit_centers,
                          marker='o', label='Single Strain')
            axs[1,1].plot(fit_E0, unconstrained_fit_centers,
                          linestyle='', marker='o', label='Unconstrained')
            axs[1,1].plot(slope * unconstrained_fit_centers + intercept,fit_E0,
                          color='C1', label='Unconstrained: Linear Fit')
            axs[1,1].legend()

            fig.tight_layout()

            if save_figures:
                figfile = os.path.join(outputdir, 'ceria_calibration_fits.png')
                plt.savefig(figfile)
                self.logger.info(f'Saved figure to {figfile}')
            if interactive:
                plt.show()

        return float(tth), float(slope), float(intercept)


class MCADataProcessor(Processor):
    """A Processor to return data from an MCA, restuctured to
    incorporate the shape & metadata associated with a map
    configuration to which the MCA data belongs, and linearly
    transformed according to the results of a ceria calibration.
    """

    def process(self, data):
        """Process configurations for a map and MCA detector(s), and
        return the calibrated MCA data collected over the map.

        :param data: input map configuration and results of ceria
            calibration
        :type data: list[dict[str,object]]
        :return: calibrated and flux-corrected MCA data
        :rtype: nexusformat.nexus.NXentry
        """

        map_config = self.get_config(
            data, 'common.models.map.MapConfig')
        calibration_config = self.get_config(
            data, 'edd.models.MCACeriaCalibrationConfig')
        nxroot = self.get_nxroot(map_config, calibration_config)

        return nxroot

    def get_nxroot(self, map_config, calibration_config):
        """Get a map of the MCA data collected by the scans in
        `map_config`. The MCA data will be calibrated and
        flux-corrected according to the parameters included in
        `calibration_config`. The data will be returned along with
        relevant metadata in the form of a NeXus structure.

        :param map_config: the map configuration
        :type map_config: MapConfig
        :param calibration_config: the calibration configuration
        :type calibration_config: MCACeriaCalibrationConfig
        :return: a map of the calibrated and flux-corrected MCA data
        :rtype: nexusformat.nexus.NXroot
        """
        # third party modules
        from nexusformat.nexus import (NXdata,
                                       NXdetector,
                                       NXinstrument,
                                       NXroot)

        # local modules
        from CHAP.common import MapProcessor

        nxroot = NXroot()

        nxroot[map_config.title] = MapProcessor.get_nxentry(map_config)
        nxentry = nxroot[map_config.title]

        for detector in calibration_config.detectors:
            nxentry.instrument = NXinstrument()
            nxentry.instrument[detector.detector_name] = NXdetector()
            nxentry.instrument[detector.detector_name].calibration = dumps(
                detector.dict())

            nxentry.instrument[detector.detector_name].data = NXdata()
            nxdata = nxentry.instrument[detector.detector_name].data
            nxdata.raw = np.empty((*map_config.shape,
                                   detector.num_bins))
            nxdata.raw.attrs['units'] = 'counts'
            nxdata.channel_energy = detector.slope_calibrated \
                * np.arange(0, detector.num_bins) \
                * (detector.max_energy_kev / detector.num_bins) \
                + detector.intercept_calibrated
            nxdata.channel_energy.attrs['units'] = 'keV'

            for map_index in np.ndindex(map_config.shape):
                scans, scan_number, scan_step_index = \
                    map_config.get_scan_step_index(map_index)
                scanparser = scans.get_scanparser(scan_number)
                nxdata.raw[map_index] = scanparser.get_detector_data(
                    calibration_config.detector_name,
                    scan_step_index)

            nxentry.data.makelink(nxdata.raw, name=detector.detector_name)
            nxentry.data.makelink(
                nxdata.channel_energy,
                name=f'{detector.detector_name}_channel_energy')
            if isinstance(nxentry.data.attrs['axes'], str):
                nxentry.data.attrs['axes'] = [
                    nxentry.data.attrs['axes'],
                    f'{detector.detector_name}_channel_energy']
            else:
                nxentry.data.attrs['axes'] += [
                    f'{detector.detector_name}_channel_energy']
            nxentry.data.attrs['signal'] = detector.detector_name

        return nxroot


class StrainAnalysisProcessor(Processor):
    """Processor that takes a map of MCA data and returns a map of
    sample strains
    """
    def process(self,
                data,
                config=None,
                save_figures=False,
                outputdir='.',
                inputdir='.',
                interactive=False):
        """Return strain analysis maps & associated metadata in an NXprocess.

        :param data: input data containing configurations for a map,
            completed ceria calibration, and parameters for strain
            analysis
        :type data: list[PipelineData]
        :param config: initialization parameters for an instance of
            CHAP.edd.models.StrainAnalysisConfig, defaults to
            None
        :type config: dict, optional
        :param save_figures: save .pngs of plots for checking inputs &
            outputs of this Processor, defaults to False
        :type save_figures: bool, optional
        :param outputdir: directory to which any output figures will
            be saved, defaults to '.'
        :type outputdir: str, optional
        :param inputdir: input directory, used only if files in the
            input configuration are not absolute paths.
            be saved, defaults to '.'
        :type inputdir: str, optional
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: NXprocess containing metadata about strain analysis
            processing parameters and empty datasets for strain maps
            to be filled in later.
        :rtype: nexusformat.nexus.NXprocess

        """
        # Get required configuration models from input data
        # map_config = self.get_config(
        #     data, 'common.models.map.MapConfig')
        ceria_calibration_config = self.get_config(
            data, 'edd.models.MCACeriaCalibrationConfig', inputdir=inputdir)
        try:
            strain_analysis_config = self.get_config(
                data, 'edd.models.StrainAnalysisConfig')
        except Exception as data_exc:
            self.logger.info('No valid strain analysis config in input '
                             + 'pipeline data, using config parameter instead')
            from CHAP.edd.models import StrainAnalysisConfig
            try:
                strain_analysis_config = StrainAnalysisConfig(
                    **config, inputdir=inputdir)
            except Exception as dict_exc:
                raise RuntimeError from dict_exc

        nxroot = self.get_nxroot(
            #map_config,
            strain_analysis_config.map_config,
            ceria_calibration_config,
            strain_analysis_config,
            save_figures=save_figures,
            outputdir=outputdir,
            interactive=interactive)
        self.logger.debug(nxroot.tree)
        return nxroot


    def get_nxroot(self,
                   map_config,
                   ceria_calibration_config,
                   strain_analysis_config,
                   save_figures=False,
                   outputdir='.',
                   interactive=False):
        """Return NXroot containing strain maps.


        :param map_config: Input map configuration
        :type map_config: CHAP.common.models.map.MapConfig
        :param ceria_calibration_config: Results of ceria calibration
        :type ceria_calibration_config:
            'CHAP.edd.models.MCACeriaCalibrationConfig'
        :param strain_analysis_config: Strain analysis processing
            configuration
        :type strain_analysis_config: CHAP.edd.models.StrainAnalysisConfig
        :param save_figures: save .pngs of plots for checking inputs &
            outputs of this Processor, defaults to False
        :type save_figures: bool, optional
        :param outputdir: directory to which any output figures will
            be saved, defaults to '.'
        :type outputdir: str, optional
        :param interactive: allow for user interactions, defaults to
            False
        :type interactive: bool, optional
        :return: NXroot containing strain maps
        :rtype: nexusformat.nexus.NXroot
        """
        from nexusformat.nexus import (NXcollection,
                                       NXdata,
                                       NXdetector,
                                       NXfield,
                                       NXparameters,
                                       NXprocess,
                                       NXroot)
        import numpy as np
        from CHAP.common import MapProcessor
        from CHAP.edd.utils import hc
        from CHAP.utils.fit import FitMap

        for detector in strain_analysis_config.detectors:
            calibration = [
                d for d in ceria_calibration_config.detectors \
                if d.detector_name == detector.detector_name][0]
            detector.add_calibration(calibration)

        nxroot = NXroot()
        nxroot[map_config.title] = MapProcessor.get_nxentry(map_config)
        nxentry = nxroot[map_config.title]
        nxroot[f'{map_config.title}_strainanalysis'] = NXprocess()
        nxprocess = nxroot[f'{map_config.title}_strainanalysis']
        nxprocess.strain_analysis_config = dumps(strain_analysis_config.dict())

        # Setup plottable data group
        nxprocess.data = NXdata()
        nxprocess.default = 'data'
        nxdata = nxprocess.data
        nxdata.attrs['axes'] = map_config.dims
        def linkdims(nxgroup):
            for dim in map_config.dims:
                nxgroup.makelink(nxentry.data[dim])
                nxgroup.attrs[f'{dim}_indices'] = \
                    nxentry.data.attrs[f'{dim}_indices']
        linkdims(nxdata)

        # Select interactive params / save figures
        if save_figures or interactive:
            import matplotlib.pyplot as plt
            from CHAP.edd.utils import select_hkls
            from CHAP.utils.general import draw_mask_1d
            for detector in strain_analysis_config.detectors:
                x = np.linspace(detector.intercept_calibrated,
                                detector.max_energy_kev \
                                * detector.slope_calibrated,
                                detector.num_bins)
                y = strain_analysis_config.mca_data(
                    detector,
                    (0,) * len(strain_analysis_config.map_config.shape))
                fig = select_hkls(detector,
                                  strain_analysis_config.materials,
                                  detector.tth_calibrated,
                                  y, x, interactive)
                if save_figures:
                    fig.savefig(os.path.join(
                        outputdir,
                        f'{detector.detector_name}_strainanalysis_hkls.png'))
                plt.close()

                if interactive:
                    self.logger.info(
                        'Interactively select a mask in the matplotlib figure')
                mask, include_bin_ranges, figure = draw_mask_1d(
                    y, xdata=x,
                    current_index_ranges=detector.include_bin_ranges,
                    label='reference spectrum',
                    title='Click and drag to select ranges of MCA data to\n'
                    + 'include when analyzing strain.',
                    xlabel='MCA channel (index)',
                    ylabel='MCA intensity (counts)',
                    test_mode=not interactive,
                    return_figure=True
                )
                detector.include_bin_ranges = include_bin_ranges
                if save_figures:
                    figure.savefig(os.path.join(
                        outputdir,
                        f'{detector.detector_name}_strainanalysis_mask.png'))
                plt.close()

            if interactive:
                from CHAP.edd.utils import select_material_params
                x = np.linspace(
                    strain_analysis_config.detectors[0].intercept_calibrated,
                    detector.max_energy_kev \
                    * detector.slope_calibrated,
                    detector.num_bins)
                y = strain_analysis_config.mca_data(
                    strain_analysis_config.detectors[0],
                    (0,) * len(strain_analysis_config.map_config.shape))
                tth = strain_analysis_config.detectors[0].tth_calibrated
                strain_analysis_config.materials = select_material_params(
                    x, y, tth, materials=strain_analysis_config.materials)

        for detector in strain_analysis_config.detectors:
            # Setup NXdata group
            self.logger.debug(
                f'Setting up NXdata group for {detector.detector_name}')
            nxprocess[detector.detector_name] = NXdetector()
            nxdetector = nxprocess[detector.detector_name]
            nxdetector.local_name = detector.detector_name
            nxdetector.detector_config = dumps(detector.dict())
            nxdetector.data = NXdata()
            det_nxdata = nxdetector.data
            det_nxdata.attrs['axes'] = map_config.dims + ['energy']
            linkdims(det_nxdata)
            all_energies = np.arange(0, detector.num_bins) \
                * (detector.max_energy_kev / detector.num_bins) \
                * detector.slope_calibrated \
                + detector.intercept_calibrated
            mask = detector.mca_mask()
            energies = all_energies[mask]
            det_nxdata.energy = NXfield(value=energies,
                                        attrs={'units': 'keV'})
            det_nxdata.attrs['energy_indices'] = len(map_config.dims)
            det_nxdata.intensity = NXfield(
                dtype='uint16',
                shape=(*map_config.shape, len(energies)),
                attrs={'units': 'counts'})
            det_nxdata.microstrain = NXfield(
                dtype='float64',
                shape=map_config.shape,
                attrs={'long_name': 'Strain (\u03BC\u03B5)'})

            # Gather detector data
            self.logger.debug(
                f'Gathering detector data for {detector.detector_name}')
            for map_index in np.ndindex(map_config.shape):
                try:
                    scans, scan_number, scan_step_index = \
                        map_config.get_scan_step_index(map_index)
                except:
                    continue
                scanparser = scans.get_scanparser(scan_number)
                intensity = scanparser.get_detector_data(
                    detector.detector_name, scan_step_index)\
                    .astype('uint16')[mask]
                det_nxdata.intensity[map_index] = intensity
            det_nxdata.summed_intensity = det_nxdata.intensity.sum(axis=-1)

            # Perform strain analysis
            self.logger.debug(
                f'Beginning strain analysis for {detector.detector_name}')
            fit_hkls, fit_ds = detector.fit_ds(
                strain_analysis_config.materials)
            peak_locations = hc / (
                2. * fit_ds * np.sin(0.5*np.radians(detector.tth_calibrated)))
            # KLS: Use the below def of peak_locations when
            # FitMap.create_multipeak_model can accept a list of maps
            # for centers.
            # tth = np.radians(detector.map_tth(map_config))
            # peak_locations = [hc / (2. * d0 * np.sin(0.5*tth)) \
            #                   for d0 in fit_ds]

            # Perform initial fit: assume uniform strain for all HKLs
            self.logger.debug('Performing uniform fit')
            fit = FitMap(det_nxdata.intensity.nxdata, x=energies)
            fit.create_multipeak_model(
                peak_locations,
                fit_type='uniform',
                peak_models=detector.peak_models,
                background=detector.background)
            fit.fit()
            uniform_fit_centers = [
                fit.best_values[
                    fit.best_parameters().index(f'peak{i+1}_center')]
                for i in range(len(peak_locations))]
            uniform_fit_errors = [
                fit.best_errors[
                   fit.best_parameters().index(f'peak{i+1}_center')]
                for i in range(len(peak_locations))]

            # Add uniform fit results to the NeXus structure
            nxdetector.uniform_fit = NXcollection()
            fit_nxgroup = nxdetector.uniform_fit
            # Full map of results
            fit_nxgroup.results = NXdata()
            fit_nxdata = fit_nxgroup.results
            fit_nxdata.attrs['axes'] = map_config.dims + ['energy']
            linkdims(fit_nxdata)
            fit_nxdata.makelink(det_nxdata.energy)
            fit_nxdata.attrs['energy_indices'] = len(map_config.dims)
            for d in fit.best_results:
                if d.endswith('_fit'):
                    fit_nxdata.fits = fit.best_results[d]
            fit_nxdata.residuals = fit.residual

            # Peak-by-peak results
            fit_nxgroup.fit_hkl_centers = NXdata()
            fit_nxdata = fit_nxgroup.fit_hkl_centers
            fit_nxdata.attrs['axes'] = map_config.dims
            linkdims(fit_nxdata)
            for hkl, center_guessed, centers_fit, centers_errors in \
                    zip(fit_hkls, peak_locations,
                        uniform_fit_centers, uniform_fit_errors):
                hkl_name = '_'.join(str(hkl)[1:-1].split(' '))
                fit_nxgroup[hkl_name] = NXparameters()
                fit_nxgroup[hkl_name].initial_guess = center_guessed
                fit_nxgroup[hkl_name].initial_guess.attrs['units'] = 'keV'
                fit_nxgroup[hkl_name].centers = NXdata()
                fit_nxgroup[hkl_name].centers.attrs['axes'] = map_config.dims
                linkdims(fit_nxgroup[hkl_name].centers)
                fit_nxgroup[hkl_name].centers.values = NXfield(
                    value=centers_fit, attrs={'units': 'keV'})
                fit_nxgroup[hkl_name].centers.errors = NXfield(
                    value=centers_errors)
                fit_nxdata.makelink(fit_nxgroup[f'{hkl_name}/centers/values'],
                                    name=hkl_name)

            # Perform second fit: do not assume uniform strain for all
            # HKLs, and use the fit peak centers from the uniform fit
            # as inital guesses
            self.logger.debug('Performing unconstrained fit')
            fit.create_multipeak_model(fit_type='unconstrained')
            fit.fit()
            unconstrained_fit_centers = np.array(
                [fit.best_values[
                    fit.best_parameters()\
                    .index(f'peak{i+1}_center')]
                 for i in range(len(peak_locations))])
            unconstrained_fit_errors = np.array(
                [fit.best_errors[
                    fit.best_parameters()\
                    .index(f'peak{i+1}_center')]
                 for i in range(len(peak_locations))])
            unconstrained_strains = np.empty_like(unconstrained_fit_centers)
            for i, peak_loc in enumerate(peak_locations):
                unconstrained_strains[i] = np.log(
                    peak_loc / unconstrained_fit_centers[i])
            unconstrained_strain = np.mean(unconstrained_strains, axis=0)
            det_nxdata.microstrain.nxdata = unconstrained_strain * 1e6

            # Add unconstrained fit results to the NeXus structure
            nxdetector.unconstrained_fit = NXcollection()
            fit_nxgroup = nxdetector.unconstrained_fit
            # Full map of results
            fit_nxgroup.data = NXdata()
            fit_nxdata = fit_nxgroup.data
            fit_nxdata.attrs['axes'] = map_config.dims + ['energy']
            linkdims(fit_nxdata)
            fit_nxdata.makelink(det_nxdata.energy)
            fit_nxdata.attrs['energy_indices'] = len(map_config.dims)
            for d in fit.best_results:
                if d.endswith('_fit'):
                    fit_nxdata.fits = fit.best_results[d]
            fit_nxdata.residuals = fit.residual
            # Peak-by-peak results
            fit_nxgroup.fit_hkl_centers = NXdata()
            fit_nxdata = fit_nxgroup.fit_hkl_centers
            fit_nxdata.attrs['axes'] = map_config.dims
            linkdims(fit_nxdata)
            for (hkl, unconstrained_center_guesses,
                 centers_fit, centers_errors) in \
                    zip(fit_hkls, peak_locations,
                        unconstrained_fit_centers, unconstrained_fit_errors):
                hkl_name = '_'.join(str(hkl)[1:-1].split(' '))
                fit_nxgroup[hkl_name] = NXparameters()
                fit_nxgroup[hkl_name].initial_guess = center_guessed
                fit_nxgroup[hkl_name].initial_guess.attrs['units'] = 'keV'
                fit_nxgroup[hkl_name].centers = NXdata()
                fit_nxgroup[hkl_name].centers.attrs['axes'] = map_config.dims
                linkdims(fit_nxgroup[hkl_name].centers)
                fit_nxgroup[hkl_name].centers.values = NXfield(
                    value=centers_fit, attrs={'units': 'keV'})
                fit_nxgroup[hkl_name].centers.errors = NXfield(
                    value=centers_errors)
                fit_nxdata.makelink(fit_nxgroup[f'{hkl_name}/centers/values'],
                                    name=hkl_name)
        return nxroot


if __name__ == '__main__':
    # local modules
    from CHAP.processor import main

    main()



"""Tomography Pydantic model classes."""

# Third party imports
from typing import (
    Literal,
    Optional,
)
from pydantic import (
    BaseModel,
    StrictBool,
    conint,
    conlist,
    confloat,
    constr,
)

# Local modules
from CHAP.common.models.map import SpecScans

class Detector(BaseModel):
    """
    Detector class to represent the detector used in the experiment.

    :ivar prefix: Prefix of the detector in the SPEC file.
    :type prefix: str
    :ivar rows: Number of pixel rows on the detector.
    :type rows: int
    :ivar columns: Number of pixel columns on the detector.
    :type columns: int
    :ivar pixel_size: Pixel size of the detector in mm.
    :type pixel_size: int or list[int]
    :ivar lens_magnification: Lens magnification for the detector,
        defaults to 1.0.
    :type lens_magnification: float, optional
    """
    prefix: constr(strip_whitespace=True, min_length=1)
    rows: conint(gt=0)
    columns: conint(gt=0)
    pixel_size: conlist(
        item_type=confloat(gt=0, allow_inf_nan=False),
        min_items=1, max_items=2)
    lens_magnification: confloat(gt=0, allow_inf_nan=False) = 1.0


class TomoReduceConfig(BaseModel):
    """
    Class representing the configuration for the tomography image
    reduction processor.

    :ivar img_row_bounds: Detector image bounds in the row-direction.
    :type img_row_bounds: list[int], optional
    :ivar delta_theta: Rotation angle increment in image reduction
        in degrees.
    :type delta_theta: float, optional
    """
    img_row_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]
    delta_theta: Optional[confloat(gt=0, allow_inf_nan=False)]


class TomoFindCenterConfig(BaseModel):
    """
    Class representing the configuration for the tomography center axis
    finding processor.

    :ivar center_stack_index: Stack index of the tomography set to find
        the center axis.
    :type center_stack_index: int, optional
    :ivar lower_row: Lower row index for the center finding processor.
    :type lower_row: int, optional
    :ivar lower_center_offset: Center at the lower row index.
    :type lower_center_offset: float, optional
    :ivar upper_row: Upper row index for the center finding processor.
    :type upper_row: int, optional
    :ivar upper_center_offset: Center at the upper row index.
    :type upper_center_offset: float, optional
    :ivar gaussian_sigma: Standard deviation for the Gaussian filter
        applied to image reconstruction visualizations, defaults to no
        filtering performed.
    :type gaussian_sigma: float, optional
    :ivar ring_width: Maximum width of rings to be filtered in the image
        reconstruction in pixels, defaults to no filtering performed.
    :type ring_width: float, optional
    """
    center_stack_index: Optional[conint(ge=0)]
    lower_row: Optional[conint(ge=-1)]
    lower_center_offset: Optional[confloat(allow_inf_nan=False)]
    upper_row: Optional[conint(ge=-1)]
    upper_center_offset: Optional[confloat(allow_inf_nan=False)]
    search_range: Optional[confloat(ge=0, allow_inf_nan=False)]
    search_step: Optional[confloat(ge=0, allow_inf_nan=False)]
    gaussian_sigma: Optional[confloat(ge=0, allow_inf_nan=False)]
    ring_width: Optional[confloat(ge=0, allow_inf_nan=False)]


class TomoReconstructConfig(BaseModel):
    """
    Class representing the configuration for the tomography image
    reconstruction processor.

    :ivar x_bounds: Reconstructed image bounds in the x-direction.
    :type x_bounds: list[int], optional
    :ivar y_bounds: Reconstructed image bounds in the y-direction.
    :type y_bounds: list[int], optional
    :ivar z_bounds: Reconstructed image bounds in the z-direction.
    :type z_bounds: list[int], optional
    :ivar secondary_iters: Number of secondary iterations in the tomopy
        image reconstruction algorithm, defaults to 0.
    :type secondary_iters: int, optional
    :ivar remove_stripe_sigma: Damping parameter in Fourier space in
        tomopy's horizontal stripe removal tool, defaults to no
        correction performed.
    :type remove_stripe_sigma: float, optional
    :ivar ring_width: Maximum width of rings to be filtered in the image
        reconstruction in pixels, defaults to no filtering performed.
    :type ring_width: float, optional
    """
    x_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]
    y_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]
    z_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]
    secondary_iters: conint(ge=0) = 0
    remove_stripe_sigma: Optional[confloat(ge=0, allow_inf_nan=False)]
    ring_width: Optional[confloat(ge=0, allow_inf_nan=False)]


class TomoCombineConfig(BaseModel):
    """
    Class representing the configuration for the combined tomography
    stacks processor.

    :ivar x_bounds: Combined image bounds in the x-direction.
    :type x_bounds: list[int], optional
    :ivar y_bounds: Combined image bounds in the y-direction.
    :type y_bounds: list[int], optional
    :ivar z_bounds: Combined image bounds in the z-direction.
    :type z_bounds: list[int], optional
    """
    x_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]
    y_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]
    z_bounds: Optional[
        conlist(item_type=conint(ge=-1), min_items=2, max_items=2)]


class TomoSimConfig(BaseModel):
    """
    Class representing the configuration for the tomography simulator.

    :ivar station: The station name (in 'idxx' format).
    :type station: Literal['id1a3', 'id3a', 'id3b']
    :ivar detector: Detector used in the tomography experiment.
    :type detector: Detector
    :ivar sample_type: Sample type for the tomography simulator.
    :type sample_type: Literal['square_rod', 'square_pipe',
        'hollow_cube', 'hollow_brick']
    :ivar sample_size: Size of each sample dimension in mm (internally
        converted to an integer number of pixels).
    :type sample_size: list[float]
    :ivar wall_thickness: Wall thickness for pipe, cube, and brick in
        mm (internally converted to an integer number of pixels).
    :type wall_thickness: float
    :ivar mu: Linear attenuation coefficient in mm^-1, defaults to 0.05.
    :type mu: float, optional
    :ivar theta_step: Rotation angle increment in the tomography
        simulation in degrees.
    :type theta_step: float
    :ivar beam_intensity: Initial beam intensity in counts,
        defaults to 1.e9.
    :type beam_intensity: float, optional
    :ivar background_intensity: Background intensity in counts,
        defaults to 20.
    :type background_intensity:: float, optional
    :ivar slit_size: Vertical beam height in mm, defaults to 1.0.
    :type slit_size:: float, optional
    """
    station: Literal['id1a3', 'id3a', 'id3b']
    detector: Detector.construct()
    sample_type: Literal[
        'square_rod', 'square_pipe', 'hollow_cube', 'hollow_brick']
    sample_size: conlist(
        item_type=confloat(gt=0, allow_inf_nan=False),
        min_items=1, max_items=2)
    wall_thickness: Optional[confloat(ge=0, allow_inf_nan=False)]
    mu: Optional[confloat(gt=0, allow_inf_nan=False)] = 0.05
    theta_step: confloat(gt=0, allow_inf_nan=False)
    beam_intensity: Optional[confloat(gt=0, allow_inf_nan=False)] = 1.e9
    background_intensity: Optional[confloat(gt=0, allow_inf_nan=False)] = 20
    slit_size: Optional[confloat(gt=0, allow_inf_nan=False)] = 1.0



"""Utility functions for EDD workflows"""

from scipy.constants import physical_constants
hc = 1e7 * physical_constants['Planck constant in eV/Hz'][0] \
     * physical_constants['speed of light in vacuum'][0]


def make_material(name, sgnum, lparms, dmin=0.6):
    """Return a hexrd.material.Material with the given properties.
    Taken from
    [CHEDDr](https://gitlab01.classe.cornell.edu/msn-c/cheddr/-/blob/master/notebooks/CHESS_EDD.py#L99).


    :param name: material name
    :type name: str
    :param sgnum: spage group number
    :type sgnum: int
    :param lparms: lattice parameters ([a, b, c, &#945;, &#946;,
        &#947;], or fewer as the symmetry of the space group allows --
        for instance, a cubic lattice with space group number 225 can
        just provide [a, ])
    :type lparms: list[float]
    :param dmin: dmin of the material in Angstrom (&#8491;), defaults to 0.6
    :type dmin: float, optional
    :return: a hexrd material
    :rtype: heard.material.Material
    """
    from hexrd.material import Material    
    from hexrd.valunits import valWUnit
    import numpy as np
    matl = Material(name=name)
    matl.sgnum = sgnum
    if isinstance(lparms, float):
        lparms = [lparms]
    matl.latticeParameters = lparms
    matl.dmin = valWUnit('lp', 'length',  dmin, 'angstrom')
    nhkls = len(matl.planeData.exclusions)
    matl.planeData.set_exclusions(np.zeros(nhkls, dtype=bool))
    return matl

def get_unique_hkls_ds(materials, tth_tol=None, tth_max=None, round_sig=8):
    """Return the unique HKLs and d-spacings for the given list of
    materials.

    :param materials: list of materials to get HKLs and d-spacings for
    :type materials: list[hexrd.material.Material]
    :return: unique HKLs, unique d-spacings
    :rtype: tuple[np.ndarray, np.ndarray]
    """
    from copy import deepcopy
    import numpy as np
    from CHAP.edd.models import MaterialConfig
    _materials = deepcopy(materials)
    for i, m in enumerate(materials):
        if isinstance(m, MaterialConfig):
            _materials[i] = m._material
    hkls = np.empty((0,3))
    ds = np.empty((0))
    ds_index = np.empty((0))
    for i, material in enumerate(_materials):
        plane_data = material.planeData
        if tth_tol is not None:
            plane_data.tThWidth = np.radians(tth_tol)
        if tth_max is not None:
            plane_data.exclusions = None
            plane_data.tThMax = np.radians(tth_max)
        hkls = np.vstack((hkls, plane_data.hkls.T))
        ds_i = plane_data.getPlaneSpacings()
        ds = np.hstack((ds, ds_i))
        ds_index = np.hstack((ds_index, i*np.ones(len(ds_i))))
    # Sort lattice spacings in reverse order (use -)
    ds_unique, ds_index_unique, _ = np.unique(
        -ds.round(round_sig), return_index=True, return_counts=True)
    ds_unique = np.abs(ds_unique)
    # Limit the list to unique lattice spacings
    hkls_unique = hkls[ds_index_unique,:].astype(int)
    ds_unique = ds[ds_index_unique]
    return hkls_unique, ds_unique


def select_hkls(detector, materials, tth, y, x, interactive):
    """Return a plot of `detector.fit_hkls` as a matplotlib
    figure. Optionally modify `detector.fit_hkls` by interacting with
    a matplotlib figure.

    :param detector: the detector to set `fit_hkls` on
    :type detector: MCAElementConfig
    :param material: the material to pick HKLs for
    :type material: MaterialConfig
    :param tth: diffraction angle two-theta
    :type tth: float
    :param y: reference y data to plot
    :type y: np.ndarray
    :param x: reference x data to plot
    :type x: np.ndarray
    :param interactive: show the plot and allow user interactions with
        the matplotlib figure
    :type interactive: bool
    :return: plot showing the user-selected HKLs
    :rtype: matplotlib.figure.Figure
    """
    import numpy as np
    hkls, ds = get_unique_hkls_ds(materials)
    peak_locations = hc / (2. * ds * np.sin(0.5 * np.radians(tth)))
    pre_selected_peak_indices = detector.fit_hkls \
                                if detector.fit_hkls else []
    from CHAP.utils.general import select_peaks
    selected_peaks, figure = select_peaks(
        y, x, peak_locations,
        peak_labels=[str(hkl)[1:-1] for hkl in hkls],
        pre_selected_peak_indices=pre_selected_peak_indices,
        mask=detector.mca_mask(),
        interactive=interactive,
        xlabel='MCA channel energy (keV)',
        ylabel='MCA intensity (counts)',
        title='Mask and HKLs for fitting')

    selected_hkl_indices = [int(np.where(peak_locations == peak)[0][0]) \
                            for peak in selected_peaks]
    detector.fit_hkls = selected_hkl_indices

    return figure


def select_tth_initial_guess(detector, material, y, x):
    """Show a matplotlib figure of a reference MCA spectrum on top of
    HKL locations. The figure includes an input field to adjust the
    initial tth guess and responds by updating the HKL locations based
    on the adjusted value of the initial tth guess.

    :param detector: the detector to set `tth_inital_guess` on
    :type detector: MCAElementConfig
    :param material: the material to show HKLs for
    :type material: MaterialConfig
    :param y: reference y data to plot
    :type y: np.ndarray
    :param x: reference x data to plot
    :type x: np.ndarray
    :return: None
    """
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Button, TextBox
    import numpy as np

    tth_initial_guess = detector.tth_initial_guess \
                        if detector.tth_initial_guess is not None \
                        else 5.0
    hkls, ds = material.unique_ds(
        tth_tol=detector.hkl_tth_tol, tth_max=detector.tth_max)
    def get_peak_locations(tth):
        return hc / (2. * ds * np.sin(0.5 * np.radians(tth)))

    fig, ax = plt.subplots(figsize=(11, 8.5))
    ax.plot(x, y)
    ax.set_xlabel('MCA channel energy (keV)')
    ax.set_ylabel('MCA intensity (counts)')
    ax.set_title('Adjust initial guess for $2\\theta$')
    hkl_lines = [ax.axvline(loc, c='k', ls='--', lw=1) \
                 for loc in get_peak_locations(tth_initial_guess)]
    hkl_lbls = [ax.text(loc, 1, str(hkl)[1:-1],
                        ha='right', va='top', rotation=90,
                        transform=ax.get_xaxis_transform()) \
                for loc, hkl in zip(get_peak_locations(tth_initial_guess),
                                    hkls)]

    # Callback for tth input
    def new_guess(tth):
        try:
            tth = float(tth)
        except:
            raise ValueError(f'Cannot convert {new_tth} to float')
        for i, (line, loc) in enumerate(zip(hkl_lines,
                                            get_peak_locations(tth))):
            line.remove()
            hkl_lines[i] = ax.axvline(loc, c='k', ls='--', lw=1)
            hkl_lbls[i].remove()
            hkl_lbls[i] = ax.text(loc, 1, str(hkls[i])[1:-1],
                                  ha='right', va='top', rotation=90,
                                  transform=ax.get_xaxis_transform())
        ax.get_figure().canvas.draw()
        detector.tth_initial_guess = tth

    # Setup tth input
    plt.subplots_adjust(bottom=0.25)
    tth_input = TextBox(plt.axes([0.125, 0.05, 0.15, 0.075]),
                        '$2\\theta$: ',
                        initial=tth_initial_guess)
    cid_update_tth = tth_input.on_submit(new_guess)

    # Setup "Confirm" button
    def confirm_selection(event):
        plt.close()
    confirm_b = Button(plt.axes([0.75, 0.05, 0.15, 0.075]), 'Confirm')
    cid_confirm = confirm_b.on_clicked(confirm_selection)

    # Show figure for user interaction
    plt.show()

    # Disconnect all widget callbacks when figure is closed
    tth_input.disconnect(cid_update_tth)
    confirm_b.disconnect(cid_confirm)



def select_material_params(x, y, tth, materials=[]):
                           # name=Material.DFLT_NAME,
                           # sgnum=Material.DFLT_SGNUM,
                           # lparms=Material.DFLT_LPARMS,
                           # dmin=0.6):
    """Interactively select lattice parameters and space group for a
    list of materials. A matplotlib figure will be shown with a plot
    of the reference data (`x` and `y`). The figure will contain
    widgets to add / remove materials and update selections for space
    group number and lattice parameters for each one. The HKLs for the
    materials defined by the widgets' values will be shown over the
    reference data and updated when the widgets' values are
    updated. Return a list of the selected materials when the figure
    is closed.
    """
    from copy import deepcopy
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Button, TextBox
    import numpy as np
    from CHAP.edd.models import MaterialConfig

    _materials = deepcopy(materials)
    for i, m in enumerate(_materials):
        if isinstance(m, MaterialConfig):
            _materials[i] = m._material

    # Set up plot of reference data
    fig, ax = plt.subplots(figsize=(11, 8.5))

    def draw_plot():
        ax.clear()
        ax.set_title('Reference Data')
        ax.set_xlabel('MCA channel energy (keV)')
        ax.set_ylabel('MCA intensity (counts)')
        ax.plot(x, y)
        for i, material in enumerate(_materials):
            hkls, ds = get_unique_hkls_ds([material])            
            E0s = hc / (2. * ds * np.sin(0.5 * np.radians(tth)))
            for hkl, E0 in zip(hkls, E0s):
                ax.axvline(E0, c=f'C{i}', ls='--', lw=1)
                ax.text(E0, 1, str(hkl)[1:-1], c=f'C{i}',
                        ha='right', va='top', rotation=90,
                        transform=ax.get_xaxis_transform())
        ax.get_figure().canvas.draw()

    # Confirm & close button
    widget_callbacks = []
    plt.subplots_adjust(bottom=0.1)
    def confirm_selection(event):
        plt.close()
    confirm_button = Button(plt.axes([0.75, 0.015, 0.1, 0.05]), 'Confirm')
    cid_confirm = confirm_button.on_clicked(confirm_selection)
    widget_callbacks.append((confirm_button, cid_confirm))

    # Widgets to edit materials
    widgets = []
    def update_materials(*args, **kwargs):
        """Validate input material properties and redraw the plot"""
        materials_ok = True
        for i, (material,
                (name_input, sgnum_input, \
                 a_input, b_input, c_input, \
                 alpha_input, beta_input, gamma_input)) \
                in enumerate(zip(materials, widgets)):
            name = name_input.text
            print(f'looking at parms for material {name}')
            try:
                sgnum = int(sgnum_input.text)
                sgnum_input.color = '.95'
                _materials[i].sgnum = sgnum
            except:
                sgnum_input.color = 'red'
                materials_ok = False
            lparms = []
            lparms_ok = True
            for lparm_input in (a_input, b_input, c_input,
                                alpha_input, beta_input, gamma_input):
                try:
                    lparm = float(lparm_input.text)
                    lparm_input.color = '.95'
                except:
                    lparm_input.color = 'red'
                    materials_ok = False
                    lparms_ok = False
                else:
                    lparms.append(lparm)
            if lparms_ok:
                _materials[i].latticeParameters = lparms


        if materials_ok:
            confirm_button.active = True
        else:
            confirm_button.active = False
        # update items in the `materials` list, then:
        draw_plot()

    def add_material(*args, material=None, new=True):
        """Add new row of material-property-editing widgets to the
        figure and update the plot with new HKLs
        """
        if material is None:
            material = make_material('new_material', 225, 3.0)
            _materials.append(material)
        elif isinstance(material, MaterialConfig):
            material = material._material
        bottom = len(_materials) * 0.075
        plt.subplots_adjust(bottom=bottom + 0.125)
        name_input = TextBox(plt.axes([0.125, bottom, 0.06, 0.05]),
                             'Material: ',
                             initial=material.name)
        sgnum_input = TextBox(plt.axes([0.3, bottom, 0.06, 0.05]),
                              'Space Group: ',
                              initial=material.sgnum)
        a_input = TextBox(plt.axes([0.4, bottom, 0.06, 0.05]),
                          '$a$ ($\\AA$): ',
                          initial=material.latticeParameters[0].value)
        b_input = TextBox(plt.axes([0.5, bottom, 0.06, 0.05]),
                          '$b$ ($\\AA$): ',
                          initial=material.latticeParameters[1].value)
        c_input = TextBox(plt.axes([0.6, bottom, 0.06, 0.05]),
                          '$c$ ($\\AA$): ',
                          initial=material.latticeParameters[2].value)
        alpha_input = TextBox(plt.axes([0.7, bottom, 0.06, 0.05]),
                              '$\\alpha$ ($\\degree$): ',
                              initial=material.latticeParameters[3].value)
        beta_input = TextBox(plt.axes([0.8, bottom, 0.06, 0.05]),
                             '$\\beta$ ($\\degree$): ',
                             initial=material.latticeParameters[4].value)
        gamma_input = TextBox(plt.axes([0.9, bottom, 0.06, 0.05]),
                              '$\\gamma$ ($\\degree$): ',
                              initial=material.latticeParameters[5].value)
        widgets.append(
            (name_input, sgnum_input, a_input, b_input, c_input,
             alpha_input, beta_input, gamma_input))
        for widget in widgets[-1]:
            widget_callbacks.append(
                (widget, widget.on_submit(update_materials)))
        draw_plot()

    # Button to add materials
    add_material_button = Button(
        plt.axes([0.125, 0.015, 0.1, 0.05]), 'Add material')
    cid_add_material = add_material_button.on_clicked(add_material)
    widget_callbacks.append((add_material_button, cid_add_material))

    # Draw data & show plot
    for material in _materials:
        add_material(material=material)
    plt.show()

    # Teardown after figure is closed
    for widget, callback in widget_callbacks:
        widget.disconnect(callback)

    new_materials = [MaterialConfig(
        material_name=m.name, sgnum=m.sgnum,
        lattice_parameters=[m.latticeParameters[i].value for i in range(6)]) \
                     for m in _materials]
    return new_materials



"""This subpackage contains `PipelineItems` unique to SAXS/WAXS data
processing workflows.
"""
# from CHAP.saxswaxs.reader import
# from CHAP.saxswaxs.processor import
# from CHAP.saxswaxs.writer import

from CHAP.common import (IntegrateMapProcessor,
                         MapProcessor)
